package de.cnc.expression;

import java.math.BigDecimal;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.cnc.expression.exceptions.ExpressionEvaluationException;
import de.cnc.expression.exceptions.ExpressionParseException;
import de.cnc.expression.infixoperators.AbstractInfixOperator;
import de.cnc.expression.tokensimple.TimeOfDayLiteral;
import de.cnc.expression.util.AbstractInfixOperatorArrayList;
import de.cnc.expression.util.Month;
import de.cnc.expression.util.ParserUtil;
import de.cnc.expression.util.SubExpressionArrayList;
import de.cnc.util.CompareUtil;
import de.cnc.util.StringUtil;

/**
 * Expression Library
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a><br/>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id: Expression.java,v 1.4 2005/05/09 08:35:17 hkuecker Exp $
 */
public class Expression extends AbstractToken implements Comparable
{
    //static boolean debugStop = false;
    
    static final String SEP_LINE =
        "-----------------------------------------------------";

    protected SubExpressionArrayList subExpressionList =
        new SubExpressionArrayList();

    protected AbstractInfixOperatorArrayList infixOperatorList =
        new AbstractInfixOperatorArrayList();

    protected OperatorPriorityLevel operatorPriorityLevel;

    /**
     * Parsen mit strikter Prfung auf Verbrauchen des gesamten Source Strings
     * mit standardmssigem Start in Zeile 0 und Spalte 0
     */
    public static Expression parse(//--
            final String paStr, //--
            final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionParseException
    {
        return parse(paStr, 0, 0, pRunEnv);
    } // end method parse

    /**
     * Parsen mit strikter Prfung auf Verbrauchen des gesamten Source Strings
     */
    public static Expression parse(//--
            final String paStr, //--
            final int paIntLine, //--
            final int paIntCol, //--
            final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionParseException
    {
        return parseStrict(paStr, paIntLine, 0, pRunEnv);
    } // end method parse

    /**
     * Parsen eines Strings der Expression Language ( ${ ... } )
     * mit standardmssigem Start in Spalte 0
     */
    public static Expression parseExpLang(//--
            final String paStr, //--
            final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionParseException
    {
        return parseExpLang(paStr, 0, 0, pRunEnv);
    } // end method parseExpLang

    /**
     * Parsen eines Strings der Expression Language ( ${ ... } )
     */
    public static Expression parseExpLang(
        String paStr,
        int paIntLine,
        int paIntCol,
        final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionParseException
    {
        String strExp = convertExpLangToExp(paStr);

        return parseStrict(strExp, paIntLine, 0, pRunEnv);
    } // end method parseExpLang

    /**
     * Parsen mit strikter Prfung auf Verbrauchen des gesamten Source Strings
     *
     * static factory method to parse string
     *
     * @param paStr string to parse
     * @param paIntLineNo the source line number
     *
     * @return an expression token
     */
    public static Expression parseStrict(
        final String paStr,
        final int paIntLine,
        final int paIntCol,
        final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionParseException
    {
        Expression exp = parseEasy(paStr, paIntLine, paIntCol, paStr);

        // System.out.println("exp.getSrcLength():" + exp.getSrcLength());

        if (exp.getSrcLength() < paStr.length())
        {
            String strRest = paStr.substring(exp.getSrcLength());

            // System.out.println("strRest:" + strRest);
            // System.out.println("strRest.length():" + strRest.length());

            int iWhites = ParserUtil.parseWhitesAndComments(strRest);

            // System.out.println("iWhites:" + iWhites);
            if (iWhites < strRest.length())
            {
                if (exp != null
                    && exp.subExpressionList != null
                    && exp.subExpressionList.size() > 0)
                {
                    // mindestens eine SubExpression gefunden
                    throw new ExpressionParseException(
                        paIntLine,
                        exp.getSrcLength() + iWhites,
                        paStr,
                        "end of expression or infix operator expected");
                }
                else
                {
                    // keine SubExpression gefunden
                    throw new ExpressionParseException(
                        paIntLine,
                        exp.getSrcLength() + iWhites,
                        paStr,
                        "no expression recognized");
                }

            }
        }

        // Klammerung der Ausdrcke entsprechend der Operator-Prioritt.
        exp.organizeOperatorPriority();


//System.out.println( exp.operatorPriorityLevel );

        //if (debugStop){
        //    int dbg = 1 ;
        //    System.out.print( dbg );
        //}
        
        // prfen der Typisierung, soweit mglich
        // hat keinen Rckgabewert, sondern wirft eine Exception
        // spter noch mit Parameter RuntimeEnvironment pRunEnv fr Variablen-Typ-Prfung
        //exp.getCheckReturnTypes();
        //muss gegen die nach Operator-Prioritt geklammerten Werte erfolgen
        exp.operatorPriorityLevel.getCheckReturnTypes(pRunEnv);

        return exp;
    } // end method

    /**
     * Parsen ohne strikte Prfung auf Verbrauchen des gesamten Source Strings . <br>
     * Ist gedacht zum Parsen in den Parameterlisten der Funktionen und Ausdruckslisten
     * mit Operatoren expression [ operator expression ].
     *
     * static factory method to parse string
     *
     * @param paStr string to parse
     * @param paIntLineNo the source line number
     *
     * @return an expression token
     */
    public static Expression parseEasy(
        String paStr,
        int paIntLine,
        int paIntCol,
        String paStrOriginalSource)
        throws ExpressionParseException
    {
        Expression exp = new Expression(paStr, paIntLine, paIntCol);

        while (true)
        {
            // linke weiss-zeichen berspringen
            int iWhites = ParserUtil.parseWhitesAndComments(paStr);

            exp.iSrcLength += iWhites;
            // System.out.println("85 exp.exp.iSrcLength "+exp.iSrcLength);

            // Expression vor/zwischen/nach Operatoren zusammen mit ihrem Prefix- und Postfix-Operator
            SubExpression subExp =
                SubExpression.parse(
                    paStr.substring(exp.iSrcLength),
                    paIntLine,
                    paIntCol + exp.iSrcLength,
                    paStrOriginalSource);

            // System.out.println("subExp:" + subExp);
            exp.subExpressionList.add(subExp);
            exp.iSrcLength += subExp.getSrcLength();

            // weiss-zeichen berspringen
            iWhites =
                ParserUtil.parseWhitesAndComments(
                    paStr.substring(exp.iSrcLength));
            exp.iSrcLength += iWhites;

            // System.out.println("89 exp.iSrcLength:"+exp.iSrcLength);
            AbstractInfixOperator infxOp =
                AbstractInfixOperator.parseInfixOperator(
                    paStr.substring(exp.iSrcLength),
                    paIntLine,
                    paIntCol + exp.iSrcLength,
                    paStrOriginalSource);

            // System.out.println("infxOp:" + infxOp);

            if (infxOp == null)
            {
                return exp;
            }
            exp.infixOperatorList.add(infxOp);
            exp.iSrcLength += infxOp.getSrcLength();

        } // while true
    } // end method parse

    /**
     * constructor
     */
    public Expression(String paStr, int paIntLine, int paIntCol)
    {
        super(paStr, paIntLine, paIntCol);
    } // end constructor

    /**
     * process the token
     */
    public Object eval(final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionEvaluationException
    {

        //    SubExpression subExp = subExpressionList.get(0);
        //
        //    Object retObject = subExp.eval(pRunEnv);
        //
        //    // Abarbeitung der Liste mit den SubExpressions und den Operatoren
        //    for (int i = 1; i < subExpressionList.size(); i++) {
        //      AbstractInfixOperator infixOperator = infixOperatorList.get(i - 1); // i- 1 es gibt immer einen Operator weniger als SubExpressions
        //
        //      subExp = subExpressionList.get(i);
        //      retObject = infixOperator.eval(retObject, subExp, pRunEnv);
        //    }// for
        //
        //    return retObject;

        //System.out.println("this: "+this);
        //System.out.println("operatorPriorityLevel: "+operatorPriorityLevel);

        return this.operatorPriorityLevel.eval(pRunEnv);
    } // end method eval

    public String toString()
    {
        //mal umbauen auf StringBuffer
        String strRet = "";
        //strRet += " <<";
        strRet += getClass().getName();

        if (this.subExpressionList != null 
                && this.subExpressionList.size() > 0)
        {
            strRet += "\n0:" + this.subExpressionList.get(0);
            // Abarbeitung der Liste mit den SubExpressions und den Operatoren
            for (int i = 1; i < this.subExpressionList.size(); i++)
            {
                strRet += "\n" + i + ":" + this.infixOperatorList.get(i - 1);
                strRet += "\n" + i + ":" + this.subExpressionList.get(i);
            } // for
        }

        //strRet += "\noperatorPriorityLevel: " + operatorPriorityLevel;
        //strRet += ">> ";
        return strRet;
    } // end method toString

    /**
     * Klammerung der Ausdrcke entsprechend der Operator-Prioritt.
     */
    public void organizeOperatorPriority()
    {
//System.out.println("organizeOperatorPriority: "+this);
//System.out.println("organizeOperatorPriority: "+this.hashCode() );
//System.out.println( ExceptionUtil.getMessageAndStackTrace( new Exception() ) );

        if (this.operatorPriorityLevel != null){
            //System.err.println(this);
            //throw new RuntimeException( "organizeOperatorPriority schon erledigt" );
            return;
        }
        
        
        this.operatorPriorityLevel =
            OperatorPriorityLevel.organizeOperatorPriority(
                this.subExpressionList,
                this.infixOperatorList);

        //System.out.println("this: "+this);
        //System.out.println("operatorPriorityLevel: "+operatorPriorityLevel);
        //@todo: eigentlich knnte hier subExpressionList und infixOperatorList gelscht werden
        //       in toString() beachten
        
        this.subExpressionList = null; // ??? nur fr Debug
        
        this.infixOperatorList = null; // ??? nur fr Debug

    }

    // prfen der Typisierung, soweit mglich
    // hat keinen Rckgabewert, sondern wirft eine Exception
    // spter noch mit Paramter RuntimeEnvironment pRunEnv fr Variablen-Typ-Prfung
    // ---- zur Zeit Aufruf nur in parseStrict, weil das die Root des Expressionbaumes ist
    // exp.infixOperator.getCheckReturnTypes(exp.leftExpression, exp.rightExpression ) ;

    /**
     * get the wahrscheinlichen return typen . <br>
     * Wahrscheinlicher Return Type zum vorbeugenden Prfen der Funktionsparameter
     * null wenn kein Rckgabe-Wert,
     * mehrere wenn mehrere mglich
     * ( z.B. Plus-Operator kann numerisch oder String liefern)
     */
    public Object[] getCheckReturnTypes(//--
            final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionParseException
    {
if (this.operatorPriorityLevel == null){
    organizeOperatorPriority();
}

        
return this.operatorPriorityLevel.getCheckReturnTypes(pRunEnv);

        
//        SubExpression subExp = subExpressionList.get(0);
//
//        int[] retTypes = subExp.getCheckReturnTypes();
//
//        // Abarbeitung der Liste mit den SubExpressions und den Operatoren
//        for (int i = 1; i < subExpressionList.size(); i++)
//        {
//System.out.println( "subExp: " + subExp );
//            AbstractInfixOperator infixOperator = infixOperatorList.get(i - 1);
//            // i- 1 es gibt immer einen Operator weniger als SubExpressions
//
//            subExp = subExpressionList.get(i);
//            int[] rightTypes = subExp.getCheckReturnTypes();
//
//System.out.println( "rightTypes: " + Types.arr2str( rightTypes ));
//
//            retTypes = infixOperator.getCheckReturnTypes(retTypes, rightTypes);
//System.out.println( "retTypes: " + Types.arr2str( retTypes ));
//        } // for
//
//        return retTypes;

    } // end method getCheckReturnTypes

    public /*SubExpression*/ AbstractToken getFirstSubExpression()
    {
if (this.operatorPriorityLevel == null){
    organizeOperatorPriority();
}

        return this.operatorPriorityLevel.getFirstSubExpression();

//        if (subExpressionList != null && subExpressionList.size() > 0)
//        {
//            return subExpressionList.get(0);
//        }
//        return null;
    }

    /**
     * method for Interface Comparable
     */
    public final int compareTo(Object paComp)
    {
        return getSource().compareTo(((Expression) paComp).getSource());
    } // end method

    /* (non-Javadoc)
     * @see de.cnc.expression.AbstractToken#isConst(de.cnc.expression.AbstractRuntimeEnvironment)
     */
    public boolean isConst(final AbstractRuntimeEnvironment pRunEnv) throws ExpressionEvaluationException
    {
//Log.debug( "isConst", this );        
        //return this.subExpressionList.isConst(pRunEnv);
        if (this.operatorPriorityLevel == null)
        {
            organizeOperatorPriority();
        }
//Log.debug( "this.operatorPriorityLevel", this.operatorPriorityLevel );        

        return this.operatorPriorityLevel.isConst(pRunEnv);
    }

    /* (non-Javadoc)
     * @see de.cnc.expression.AbstractToken#getConstVal(de.cnc.expression.AbstractRuntimeEnvironment)
     */
    public Object getConstVal(final AbstractRuntimeEnvironment pRunEnv) throws ExpressionEvaluationException
    {
//TODO Aufruf eval ausser fr Sequence in Oberklasse bndeln        
        return eval(pRunEnv);
    }


    
    // ************************************* Java Standard Tag Library  Expression Language Style Converter *************************************

    /**
     * konvertiert einen String mit der Notation der JSP-Taglib-Expression-Language in
     * einen String der de.cnc.expression.Expression-Notation . <br>
     * abc${var1}def => 'abc'+var1+'def'  <br>
     * Achtung: in der Expression drfen keine schliessenden geschwungenen Klammern
     * auftauchen. Es erfolgt keine Analyse der Expression.
     */
    public static String convertExpLangToExp(String paStrExpLang)
    {
        StringBuffer strBuff = new StringBuffer("'"); // beginnendes HochKomma

        for (int i = 0; i < paStrExpLang.length(); i++)
        {
            // System.out.println( "paStrExpLang.indexOf(\"${\", i):"+paStrExpLang.indexOf("${", i) ) ;
            // System.out.println( "paStrExpLang.substring( i ):"+paStrExpLang.substring( i ) ) ;
            if (!(paStrExpLang.indexOf("${", i) == i))
            {
                // keine Expression
                strBuff.append(paStrExpLang.charAt(i));
            }
            else
            {
                strBuff.append("'+");
                i += 2;
                for (;
                    i < paStrExpLang.length() && paStrExpLang.charAt(i) != '}';
                    i++)
                {
                    strBuff.append(paStrExpLang.charAt(i));
                }
                strBuff.append("+'");

            }
        }
        strBuff.append("'"); // endendes Hochkomma
        return strBuff.toString();
    } // end method convertExpLangToExp

    /**
     * Kompilieren eines Template-Strings in einen Expression-String . <br>
     * Analog zu den JSP-Tags
     * <pre>
     *   &lt;% ...scriptlet code... %&gt;
     * 
     *   &lt;%= ...scriptlet expression... %&gt;
     * </pre>
     * 
     * 
     * @param tmplStr
     * @return
     */
    public static String compileTemplate(String tmplStr)
    {
        //open sequence
        StringBuffer strBuff = new StringBuffer("{CNC_TMPL_OUT:='");
        for (int i = 0; i < tmplStr.length();)
        {
            if (tmplStr.startsWith("<%=", i))
            {
                strBuff.append("';");

                strBuff.append(";CNC_TMPL_OUT+=''+");
                i += 3;
                while (i < tmplStr.length() && (!tmplStr.startsWith("%>", i)))
                {
                    strBuff.append(tmplStr.charAt(i));
                    i++;
                }
                i += 2; // %> berspringen
                strBuff.append(";CNC_TMPL_OUT+='");
            }
            else if (tmplStr.startsWith("<%", i))
            {
                strBuff.append("';");
                i += 2;
                while (i < tmplStr.length() && (!tmplStr.startsWith("%>", i)))
                {
                    strBuff.append(tmplStr.charAt(i));
                    i++;
                }
                i += 2; // %> berspringen
                strBuff.append("CNC_TMPL_OUT+='");
            }
            else if (tmplStr.startsWith("'", i))
            {
                strBuff.append("'+highcomma+'"); // Encoding Hochkomma
                i++;
            }
            else
            {
                strBuff.append(tmplStr.charAt(i));
                i++;
            }
        }
        // close sequence
        // return the concatenated string
        return strBuff.toString() + "';CNC_TMPL_OUT}";
    }

    // *****************************************************************************************************************

    public static void testParse(String paStr)
    {
        System.out.println(SEP_LINE);

        try
        {
            System.out.println(paStr);
            System.out.println("" + parseStrict(paStr, 0, 0, null));
        }
        catch (ExpressionParseException exc)
        {
            System.err.println(exc.getMessage());
            exc.printStackTrace();
            throw new RuntimeException("Test parse failed");
        }

    } // end method testParse

    /**
     * Testen auf Parser-Fehlermeldung
     *
     * @param pStr Source String
     * @param pColPos Spaltenpostion, an welcher der Fehlergemeldetwerdensollte
     */
    public static void testParseException(//--
            final String pStr, //--
            final int pColPos)
    {
        testParseException( pStr, (AbstractRuntimeEnvironment) null, pColPos );
    }
    /**
     * Testen auf Parser-Fehlermeldung
     *
     * @param pStr Source String
     * @param pColPos Spaltenpostion, an welcher der Fehlergemeldetwerdensollte
     */
    public static void testParseException(//--
            final String pStr, //--
            final AbstractRuntimeEnvironment pRunEnv, //--
            final int pColPos)
    {
        System.out.println(SEP_LINE);
        boolean bExcRaised = false;

        try
        {
            System.out.println(pStr);
            // System.out.println("" + parseStrict(paStr, 0, 0));
            System.out.println("" + parse(pStr, pRunEnv));
        }
        catch (ExpressionParseException exc)
        {
            System.err.println(exc.getMessage());
            if (exc.getCol() == pColPos)
            {
                bExcRaised = true;
            }
            else
            {
                throw new RuntimeException(
                    "Test failed,wrong position " + exc.getCol() + ", expected was " + pColPos);
            }
        }
        if (!bExcRaised)
        {
            throw new RuntimeException("Test failed, no ParseException throwed");
        }
    } // end method testParse

    /**
     *
     * @param paObjOk erwarteter Wert
     */
    public static void testEval(String paStr, Object paObjOk)
    {
        testEval(paStr, new StandaloneRuntimeEnvironment(), paObjOk);
    } // end method testEval

    public static void testExpLangEval(String paStr, Object paObjOk)
    {
        testEval(convertExpLangToExp(paStr), paObjOk);
    } // end method testExpLangEval

    /**
     *
     * @param pObjOk erwarteter Wert
     */
    public static void testEval(
        String paStr,
        AbstractRuntimeEnvironment pRunEnv,
        Object pObjOk)
    {
        System.out.println(
            "-----------------------------------------------------");

        // Laufzeit-Umgebung mit Variablen
        // RuntimeEnvironment runEnv = new RuntimeEnvironment();
        Object objEval = null;

        try
        {
            System.out.println(paStr);
            objEval = parseStrict(paStr, 0, 0, pRunEnv).eval(pRunEnv);
            System.out.println("" + objEval);

        }
        catch (ExpressionParseException exc)
        {
            System.err.println(exc.getMessage());
            exc.printStackTrace();
            throw new RuntimeException("Test failed");
        }
        catch (ExpressionEvaluationException evalExc)
        {
            System.err.println(evalExc.getMessage());
            evalExc.printStackTrace();
            throw new RuntimeException("Test failed");
        }
if (pObjOk instanceof Date && objEval instanceof Date){
 System.out.println("pObjOk: " + pObjOk + " " + ( (Date) pObjOk ).getTime() );   
 System.out.println("objEval: " + objEval + " " + ( (Date) objEval ).getTime() );   
}
        if (pObjOk != null)
        {
            if (objEval == null)
            {
                throw new RuntimeException(
                    "Test failed, value is "
                        + objEval
                        + " ,expected was "
                        + pObjOk);
            }
            // Vergleich Klassen
            if (!(pObjOk.getClass().isInstance(objEval)))
            {
                throw new RuntimeException(
                    "Test failed, value is "
                        + objEval //--
                        + "(" + getClassName( objEval ) + ")" //--
                        + " ,expected was "
                        + pObjOk
                        + "(" + getClassName( pObjOk ) + ")" );
            }
            // Sonderbehandlung Boolean
            if (pObjOk instanceof Boolean)
            {
                if (!pObjOk.equals(objEval))
                {
                    throw new RuntimeException("Test failed, value not ok");
                }
            }

            // Wert vergleichen
            else if ( pObjOk instanceof Comparable )
            {
                if ( ( (Comparable) pObjOk).compareTo(objEval) != 0 )
                {
                    // throw new RuntimeException("Test failed, value not ok");
                    throw new RuntimeException(
                            "Test failed, value is "
                            + objEval //--
                            + "(" + getClassName( objEval ) + ")" //--
                            + " ,expected was "
                            + pObjOk
                            + "(" + getClassName( pObjOk ) + ")" );
                }
            }
            else if ( ! CompareUtil.isEqual( pObjOk , objEval ) )
            {
                // throw new RuntimeException("Test failed, value not ok");
                throw new RuntimeException(
                        "Test failed, value is "
                        + objEval //--
                        + "(" + getClassName( objEval ) + ")" //--
                        + " ,expected was "
                        + pObjOk
                        + "(" + getClassName( pObjOk ) + ")" );
            }
        }
        else
        {
            // Vergleich ob erwarteter und errechneter Wert gleich null
            if (objEval != null)
            {
                throw new RuntimeException(
                        "Test failed, value is "
                        + objEval //--
                        + "(" + getClassName( objEval ) + ")" //--
                        + " ,expected was "
                        + pObjOk
                        + "(" + getClassName( pObjOk ) + ")" );
            }
        }
    } // end method testEval

    private static String getClassName(final Object pObj)
    {
        if (pObj == null)
        {
            return null;
        }
        return pObj.getClass().getName();
    }

    /**
     * Testen auf Evaluation-Fehlermeldung
     *
     * @param paStr Source String
     * @param iPaColPos Spaltenpostion, an welcher der Fehlergemeldetwerdensollte
     */
    public static void testEvalException(String paStr, int iPaColPos)
    {
        // Laufzeit-Umgebung mit Variablen
        AbstractRuntimeEnvironment runEnv = new StandaloneRuntimeEnvironment();

        testEvalException(paStr, runEnv, iPaColPos);
    } // end method testEvalException

    /**
     * Testen auf Evaluation-Fehlermeldung
     *
     * @param paStr Source String
     * @param iPaColPos Spaltenpostion, an welcher der Fehlergemeldetwerdensollte
     */
    public static void testEvalException(
        String paStr,
        AbstractRuntimeEnvironment pRunEnv,
        int iPaColPos)
    {
        System.out.println(
            "-----------------------------------------------------");

        boolean bExcRaised = false;

        try
        {
            System.out.println(paStr);
            System.out.println("" + parseStrict(paStr, 0, 0, pRunEnv).eval(pRunEnv));

        }
        catch (ExpressionParseException exc)
        {
            System.err.println(exc.getMessage());
            exc.printStackTrace();
            throw new RuntimeException("Test failed");
        }
        catch (ExpressionEvaluationException evalExc)
        {
            System.err.println(evalExc.getMessage());
            evalExc.printStackTrace();
            if (evalExc.getCol() == iPaColPos)
            {
                bExcRaised = true;
            }
            else
            {
                throw new RuntimeException(
                    "Test failed,wrong position "
                        + evalExc.getCol()
                        + ", expected was "
                        + iPaColPos);
            }
        }
        if (!bExcRaised)
        {
            throw new RuntimeException("Test failed");
        }
    } // end method testEvalException

    /**
     * The Main
     */
    public static void main(String[] args)
    {
        try
        {
            testParse("\"abc\"");
            testParse("   \"abc\"   ");
            testParseException("   \"abc\"   ?", 11);
            testParse("\"abc\\\\def\"");
            testParseException("\"abc\\def\"", 4);
            testParse("123");
            testParse("123.45");
            testParse(".123");
            testParse("0.123");
            testParse("-123");
            testParse("-1.23");
            testParse("12-3");
            testParseException("1.2.3", 3);
            testParse("  12.3  ");
            testParse("true");
            testParse("false");
            testParse("null");
            testParse("var1");
            // testParse(" set() ");
            // testParse(" set(  ) ");
            // testParse(" set( 123 ) ");
            // testParse(" set( func2( 123 ) ) ");
            // testParseException(" set( func2( 123 )  ", 21);
            // testParse(" ( set( func2( 123 ) ) ) ");
            // testParse(" ( set( func2( 123 ) ) ) ");
            // testParse(" ( set( func2( 123 , \"abc\" , true ) ) ) ");
            testParse("1+2");
            testParse("1-2");
            testParseException("1-+2", 2);
            testParseException("1+\"def\"", 1);
            testParseException("null + 33", 5);
            testParseException("null - 33", 5);
            testParseException("das ist falsch", 4);

            testParse("\"abc\"+apostroph+\"def\"");

            testParse("'abc'+apostroph+'def'");

            testParse("varArr[ 0 ]");

            System.out.println(
                "-----------------------------------------------------");

            // Laufzeit-Umgebung mit Variablen
            AbstractRuntimeEnvironment runEnv =
                new StandaloneRuntimeEnvironment();

            // Setzen Anfangsvariable
            runEnv.setVariable("a", new BigDecimal(1));
            runEnv.setVariable("b", new BigDecimal(2));

            try
            {
                System.out.print("1+2 -> ");
                Expression exp1 = parseStrict("1+2", 0, 0, runEnv);

                System.out.println("" + exp1);
                try
                {

                    Object oValue = null;

                    oValue = exp1.eval(runEnv);
                    System.out.println("oValue:" + oValue);

                }
                catch (ExpressionEvaluationException evalExc)
                {
                    System.err.println(evalExc.getMessage());
                }

            }
            catch (ExpressionParseException parExc)
            {
                System.err.println(parExc.getMessage());
            }

            System.out.println(
                "-----------------------------------------------------");

            testEval("true", Boolean.TRUE);

            testEval("false", Boolean.FALSE);

            testEval("null", null);

            testEval("1+2", new BigDecimal(3));

            testEval("1*2", new BigDecimal(2));

            testEval("5**2", new BigDecimal(25));

            testEval("5 ** 2", new BigDecimal(25));

            testEval("5^2", new BigDecimal(25));

            testEval("2 ^ 5", new BigDecimal(32));

            testEval("1-2", new BigDecimal(-1));

            testEval("1/2", new BigDecimal(0.5));

            testEval("23%10", new BigDecimal(3));

            testEval("23 mod 10", new BigDecimal(3));

            testEval("\"abc\"+\"def\"", "abcdef");

            testEval("\"abc\"+2", "abc2");

            testEval("\"abc\"+2.1", "abc2.1");

            testEval("1==2", Boolean.FALSE);

            testEval("12==12", Boolean.TRUE);

            testEval("1!=2", Boolean.TRUE);

            testEval("1<>2", Boolean.TRUE);

            testEval("12!=12", Boolean.FALSE);

            testEval("12<>12", Boolean.FALSE);

            testEval("1<=2", Boolean.TRUE);

            testEval("12<=12", Boolean.TRUE);

            testEval("1>=2", Boolean.FALSE);

            testEval("12>=12", Boolean.TRUE);

            testEval("1<2", Boolean.TRUE);

            testEval("12<12", Boolean.FALSE);

            testEval("1>2", Boolean.FALSE);

            testEval("12>12", Boolean.FALSE);

            testEval("12>10", Boolean.TRUE);

            testEval("\"abc\">\"def\"", Boolean.FALSE);

            testEval("\"abc \"+true", "abc true");

            testEval("false>true", Boolean.FALSE);

            testEval("false>=true", Boolean.FALSE);

            testEval("false<=true", Boolean.TRUE);

            testEval("false!=true", Boolean.TRUE);

            testEval("false==true", Boolean.FALSE);

            testEval("( 2 + 3 ) + ( 4 + 5 )", new BigDecimal(14));

            testEvalException("3>\"def\"", 1);

            testEvalException("3>'def'", 1);

            testEval("'abc'>'def'", Boolean.FALSE);

            testEval("'abc'<'def'", Boolean.TRUE);

            testEval("( 2 + 3 ) > ( 4 + 5 )", Boolean.FALSE);

            testEval("true && false", Boolean.FALSE);

            testEval("true || false", Boolean.TRUE);

            testEval(
                "( ( 2 + 3 ) > ( 4 + 5 ) ) && ( ( 2 + 3 ) > ( 4 + 5 ) )",
                Boolean.FALSE);

            testEval(
                "( ( 2 + 3 ) > ( 4 + 5 ) ) || ( ( 2 + 3 ) > ( 4 + 5 ) )",
                Boolean.FALSE);

            // Laufzeit-Umgebung mit Variablen
            AbstractRuntimeEnvironment runEnv1 =
                new StandaloneRuntimeEnvironment();

            // Setzen Anfangsvariable
            runEnv1.setVariable("var1", new BigDecimal(1));
            runEnv1.setVariable("var2", new BigDecimal(2));
            runEnv1.setVariable("var3", Boolean.TRUE);
            System.out.println("runEnv1 " + runEnv1);
            testEval("var1", runEnv1, new BigDecimal(1));

            testEval("var1+var2", runEnv1, new BigDecimal(3));

            testEval("- 1", runEnv1, new BigDecimal(-1));

            testEval("! true", runEnv1, Boolean.FALSE);

            testEval("var1+ -var2", runEnv1, new BigDecimal(-1));

            testEval("-var1 + -var2", runEnv1, new BigDecimal(-3));

            testEval("-var1 + var2", runEnv1, new BigDecimal(1));

            testEval("!var3", runEnv1, Boolean.FALSE);

            testEval("true and false", Boolean.FALSE);

            testEval("true or false", Boolean.TRUE);

            testEval("true & false", Boolean.FALSE);

            testEval("true | false", Boolean.TRUE);

            testEval("true nand false", Boolean.TRUE);

            testEval("true nor false", Boolean.FALSE);

            testEval("true xor false", Boolean.TRUE);

            // test verkrzte ausfhrung mit nicht existierender variable
            testEval("true or novar1", Boolean.TRUE);
            testEval("true nor novar1", Boolean.FALSE);

            // test verkrzte ausfhrung mit nicht existierender variable
            testEval("false and novar1", Boolean.FALSE);
            testEval("false nand novar1", Boolean.TRUE);

            // test verkrzte ausfhrung mit nicht existierender variable
            testEval("true || novar1", Boolean.TRUE);

            // test verkrzte ausfhrung mit nicht existierender variable
            testEval("false && novar1", Boolean.FALSE);

            runEnv1.setVariable("int1", new Integer(1));

            runEnv1.setVariable("big2", new BigDecimal(2));

            runEnv1.setVariable("i", new BigDecimal(1));
            System.out.println("runEnv1:" + runEnv1);
            testEval("i", runEnv1, new BigDecimal(1));

            testEval("++i", runEnv1, new BigDecimal(2));

            testEval("--i", runEnv1, new BigDecimal(1));

            // ich weiss nicht, ob ich das nicht lieber verbieten sollte
            testEval("++i--", runEnv1, new BigDecimal(2));

            testEval("i", runEnv1, new BigDecimal(1));

            // Division durch null abfangen, auch bei modulo
            testEvalException("1/0", 1);
            testEvalException("23%0", 2);

            testEvalException("23 mod 0", 3);

            testEval(
                "setVarForName( \"set1\" , 1 )",
                runEnv1,
                new BigDecimal(1));
            // Aufruf set-Funktion
            testEval("\"set1:\"+set1", runEnv1, "set1:1");
            // Test, ob Setzen erfolgreich war

            testEval("substring(\"abc\" , 1 )", "bc");

            testEval("substring(\"abc\" , 1 , 2 )", "b");

            testEval("startsWith(\"abc\" , \"a\" )", Boolean.TRUE);

            testEval("startsWith(\"abc\" , \"x\" )", Boolean.FALSE);

            testEval("startsWith(\"abc\" , \"a\" , 0 )", Boolean.TRUE);

            testEval("startsWith(\"abc\" , \"x\" , 0 )", Boolean.FALSE);

            testParseException("startsWith()", 0); // check param count

            testParseException("substring()", 0); // check param count

            testEval("endsWith(\"abc\" , \"c\" )", Boolean.TRUE);

            testEval("endsWith(\"abc\" , \"x\" )", Boolean.FALSE);

            testEval("upper(\"abc\")", "ABC");

            testEval("lower(\"ABC\")", "abc");

            testEval("length(\"ABC\")", new BigDecimal(3));

            testEval("chr( 65 )", "A");

            testEval("char( 65 )", "A");

            testEval("charCode(\"A\")", new BigDecimal(65));

            testEval("asc(\"A\")", new BigDecimal(65));

            testEval("left(\"ABC\",1)", "A");

            testEval("left(\"ABC\",10)", "ABC");

            testEval("right(\"ABC\",1)", "C");

            testEval("right(\"ABC\",10)", "ABC");

            testEval("contents(\"ABC\",\"A\")", Boolean.TRUE);

            testEval("contents(\"ABC\",\"A\",1)", Boolean.FALSE);

            testEval("contentsIgnoreCase(\"ABC\",\"a\")", Boolean.TRUE);

            testEval("contentsIgnoreCase(\"ABC\",\"a\",1)", Boolean.FALSE);

            testEval("\"abc\"+apostroph+\"def\"", "abc\"def");

            testEval("\"abc\"+quote+\"def\"", "abc\"def");

            testEval("\"abc\"+highcomma+\"def\"", "abc'def");

            testEval("\"abc\"+highComma+\"def\"", "abc'def");

            testEval("'abc'+apostroph+'def'", "abc\"def");

            testEval("'abc'+highcomma+'def'", "abc'def");

            testEval("'abc'+highComma+'def'", "abc'def");

            testEval("'abc'+newline+'def'", "abc\ndef");

            testEval("'abc'+newLine+'def'", "abc\ndef");

            testEval("pad('abc',5)+'.'", "abc  .");

            testEval("padr('abc',2)+'.'", "abc.");

            // System.out.println("" + (1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 - 10 - 9 - 8 - 7 - 6 - 5 - 4 - 3 - 2 - 1));
            // testParse("1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 - 10 - 9 - 8 - 7 - 6 - 5 - 4 - 3 - 2 - 1 ");

            testEval(
                "1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 - 10 - 9 - 8 - 7 - 6 - 5 - 4 - 3 - 2 - 1 ",
                new BigDecimal(0));

            testEval(
                "1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10 + - 10 + - 9 + - 8 + - 7 + - 6 + - 5 + - 4 + - 3 + - 2 + - 1 ",
                new BigDecimal(0));

            testParse("1 - - 2");

            testEval(
                "1 + ( 2 + ( 3 + ( 4 + ( 5 + ( 6 + ( 7 + ( 8 + ( 9 + ( 10 - 10 - 9 - 8 - 7 - 6 - 5 - 4 - 3 - 2 - 1 )))))))))",
                new BigDecimal(0));

            testEval("1 + ( 2 - 1 ) - 2", new BigDecimal(0));

            testEval("padl('abc',5)+'.'", "  abc.");

            testEval("'.'+spaces(3)+'.'", ".   .");

            System.out.println(
                "-----------------------------------------------------");

            String strExpLang = "abc${var1}def";

            System.out.println(strExpLang);
            System.out.println(convertExpLangToExp(strExpLang));

            strExpLang = "abc${var1}def${var2}ghi";
            System.out.println(strExpLang);
            System.out.println(convertExpLangToExp(strExpLang));

            strExpLang = "${var1}def";
            System.out.println(strExpLang);
            System.out.println(convertExpLangToExp(strExpLang));

            strExpLang = "abc${var1}";
            System.out.println(strExpLang);
            System.out.println(convertExpLangToExp(strExpLang));

            // System.out.println("-----------------------------------------------------");

            testExpLangEval("abc${1}def${2}ghi", "abc1def2ghi");

            testEval("1 - - 2", new BigDecimal(3));

            testEval("str(3)", "3");

            testEval("str ( 3 ) ", "3");

            testEval("sqrt(9)", new BigDecimal(3));

            testEval("max( 3 , 2 )", new BigDecimal(3));

            testEval("min( 3 , 2 )", new BigDecimal(2));

            testEval("reciproc( 4 )", new BigDecimal(1.0 / 4.0));

            testEval("isNull( null )", Boolean.TRUE);

            testEval("isNull( 1 )", Boolean.FALSE);

            runEnv1.setVariable(
                "varArr1",
                new String[] { "varArr 0", "varArr 1", "varArr 2" });
            System.out.println("runEnv1 " + runEnv1);

            testEval("varArr1[ 0 ]", runEnv1, "varArr 0");
            testEval("varArr1[ 1 ]", runEnv1, "varArr 1");
            testEval("varArr1[ 2 ]", runEnv1, "varArr 2");
            testEvalException("varArr1[ 4 ]", runEnv1, 9);
            testEvalException("varArr1[ -1 ]", runEnv1, 9);
            testEvalException("novarArr1[ 4 ]", runEnv1, 11);

            List testList = new ArrayList();

            testList.add("varList 0");
            testList.add("varList 1");
            testList.add("varList 2");

            runEnv1.setVariable("varList", testList);
            testEval("varList[ 0 ]", runEnv1, "varList 0");
            testEval("varList[ 1 ]", runEnv1, "varList 1");
            testEval("varList[ 2 ]", runEnv1, "varList 2");
            testEvalException("varList[ 4 ]", runEnv1, 9);

            runEnv1.setVariable("index", new Integer(0));
            for (int i = 0; i < 3; i++)
            {
                testEval("varList[" + i + "]", runEnv1, "varList " + i);
            }
            for (int i = 0; i < 3; i++)
            {
                testEval("varList[index++]", runEnv1, "varList " + i);
            }
            for (int i = 2; i >= 0; i--)
            {
                testEval("varList[--index]", runEnv1, "varList " + i);
            }

            ArrayList test2ArrayList = new ArrayList();

            test2ArrayList.add(new Integer(0));
            test2ArrayList.add(new Integer(1));
            test2ArrayList.add(new Integer(2));

            runEnv1.setVariable("varvarList2", test2ArrayList);
            System.out.println("runEnv1 " + runEnv1);

            testEval("++varvarList2[0]", runEnv1, new BigDecimal(1));

            testEval("varvarList2[0]--", runEnv1, new BigDecimal(1));

            testEval("varvarList2[0]", runEnv1, new BigDecimal(0)); // new Integer(0));

            runEnv1.setVariable("varArr2", new int[] { 1 });
            System.out.println("runEnv1 " + runEnv1);

            testEval("++varArr2[0]", runEnv1, new BigDecimal(2));

            testEval("varArr2[0]", runEnv1, new Integer(2));

            testEval("if( true , 1 , 2 )", new BigDecimal(1));

            testEval("if( false , 1 , 2 )", new BigDecimal(2));

            testEval("iif( true , 1 , 2 )", new BigDecimal(1));

            testEval("iif( false , 1 , 2 )", new BigDecimal(2));

            testEval("isBool( true )", Boolean.TRUE);

            testEval("isBool( 1 )", Boolean.FALSE);

            testEval("isBoolean( true )", Boolean.TRUE);

            testEval("isBoolean( 1 )", Boolean.FALSE);

            testEval("isNum( 1 )", Boolean.TRUE);

            testEval("isNum( true )", Boolean.FALSE);

            testEval("isNumber( 1 )", Boolean.TRUE);

            testEval("isNumber( true )", Boolean.FALSE);

            testEval("isStr( 'a' )", Boolean.TRUE);

            testEval("isStr( true )", Boolean.FALSE);

            testEval("isString( 'a' )", Boolean.TRUE);

            testEval("isString( true )", Boolean.FALSE);

            testEval("round( 33.1 )", new BigDecimal(33));

            // System.out.println( "BigDecimal( 33.1 ):"+( new BigDecimal( "33.1" ) ) ) ;
            testEval("round( 33.123 , 1 )", new BigDecimal("33.1"));

            testEvalException("33 + round( 33.123 , -1 )", 5);
            // soll schief gehen, Runden auf -1 Stellen nicht erlaubt


            runEnv.setVariable("varArr2", new int[1] );
            testEval( "varArr2[0] := 10", runEnv, new BigDecimal(10));
                System.out.println( "runEnv: " + runEnv);

                // Test, ob Setzen erfolgreich war
                testEval("'varArr2[0]:'+varArr2[0]", runEnv, "varArr2[0]:10");

            // Aufruf set-Funktion
            testEval(
                "setVarForName( 'varArr2[0]' , 10 )",
                runEnv1,
                new BigDecimal(10));
            System.out.println( "runEnv1: " + runEnv1);

            // Test, ob Setzen erfolgreich war
            testEval("'varArr2[0]:'+varArr2[0]", runEnv1, "varArr2[0]:10");

            testEval(
                "setVarForName( 'varArr2[1 * 1 + 1 * -1]' , 10 )",
                runEnv1,
                new BigDecimal(10));
            // Aufruf set-Funktion mit Test Operator-Priority
            testEval(
                "'varArr2[0]:'+varArr2[2 * 1 + 2 * -1]",
                runEnv1,
                "varArr2[0]:10");
            // Test, ob Setzen erfolgreich war

            runEnv1.setVariable("varArr3", new int[] { 0 });
            System.out.println("runEnv1 " + runEnv1);

            runEnv1.setVariable("var3", new Integer(0));
            System.out.println("runEnv1 " + runEnv1);

            testEval(
                "setVarForName( 'varArr3[0]' , 11 )",
                runEnv1,
                new BigDecimal(11));
            // Aufruf set-Funktion
            testEval("'varArr3[0]:'+varArr3[0]", runEnv1, "varArr3[0]:11");
            // Test, ob Setzen erfolgreich war

            testEval("setVar( varArr3[0] , 12 )", runEnv1, new BigDecimal(12));
            // Aufruf set-Funktion
            testEval("'varArr3[0]:'+varArr3[0]", runEnv1, "varArr3[0]:12");
            // Test, ob Setzen erfolgreich war

            testEval("setVar( var4 , 13 )", runEnv1, new BigDecimal(13));
            // Aufruf set-Funktion
            testEval("'var4:'+var4", runEnv1, "var4:13");
            // Test, ob Setzen erfolgreich war

            runEnv1.setVariable("a", new Integer(0));
            System.out.println("runEnv1 " + runEnv1);

            testEval("setVar( b , ++a++ )", runEnv1, new BigDecimal(1));

            testEval("a", runEnv1, new BigDecimal(2));

            testEval("println( 1 )", new BigDecimal(1));

            // \\n es wird zweimal decodiert, einmal vom javac und nochmal von der Expression-Lib
            // evtl. mal ndern
            testEval("print( '.print + \\\\n \n.' )", ".print + \\n \n.");

            testEval(
                "{ println( 1 ) ; println( 2 ) ; println( 3 ) }",
                new BigDecimal(3));

            testEval(
                "if( true , { println( 1 ) ; println( 2 ) ; println( 3 ) },\n"
                    + "{ println( 4 ) ; println( 5 ) ; println( 6 ) } )\n",
                new BigDecimal(3));

            testEval(
                "if( false , { println( 1 ) ; println( 2 ) ; println( 3 ) },\n"
                    + "{ println( 4 ) ; println( 5 ) ; println( 6 ) } )\n",
                new BigDecimal(6));

            runEnv1.setVariable("i", new Integer(0));
            testEval("i < 3", runEnv1, Boolean.TRUE);
            testEval(
                "while( ( i <= 3 ) , { print( 'while:' ) ; println( i++ ) } )",
                runEnv1,
                new BigDecimal(3));

            runEnv1.setVariable("i", new BigDecimal(0));
            testEval(
                "repeat( ( i > 3 ) , { print( 'repeat:' ) ; println( i ) } )",
                runEnv1,
                new BigDecimal(0));

            runEnv1.setVariable("i", new BigDecimal(0));
            testEval(
                "repeat( i > 3 , { print( 'repeat:' ) ; println( i ) } )",
                runEnv1,
                new BigDecimal(0));

            testEval(
                "for( setVar( i , 0 ) , i <= 3 , i++ , { print( 'for:' ) ; println( i ) } )",
                runEnv1,
                new BigDecimal(3));

            testEval(
                "for( i := 0 , i <= 3 , i++ , { print( 'for:' ) ; println( i ) } )",
                runEnv1,
                new BigDecimal(3));

            testEval(
                "{ println( 1 ) ; println( 2 ) ; breakSequ( 3 ) ; println( 4 ) }",
                new BigDecimal(3));

            testEval(
                "{ setVar( i , 0 ) ; { println( 'sequ:'+i++ ) ; if( i < 3 , contSequ() , breakSequ( i ) ) } }",
                new BigDecimal(3));

            testEval(
                "{ setVar( i , 0 ) ; { println( 'sequ:'+i++ ) ; if( i < 3 , contSequ() , breakSequ( i ) ) } }",
                new BigDecimal(3));

            testEval("indexOf( 'abc' , 'b' )", new BigDecimal(1));

            testEval("indexOf( 'xyz' , 'b' )", new BigDecimal(-1));

            testEval("lastIndexOf( 'abc abc' , 'b' )", new BigDecimal(5));

            testEval("lastIndexOf( 'xyz' , 'b' )", new BigDecimal(-1));

            try
            {
                testEval("compile( '1' )", parse("1", null));
            }
            catch (Exception e)
            {
                e.printStackTrace();
                throw new RuntimeException("" + e);
            }

            testEval("setVar( i , 1 )", runEnv1, new BigDecimal(1));
            testEval("isExpression( i )", runEnv1, Boolean.FALSE);

            testEval("isExpression( compile( '1' ) )", Boolean.TRUE);

            testEval("eval( compile( '1' ) )", new BigDecimal(1));

            testEval("eval( '1' )", new BigDecimal(1));

            try
            {
                testEval(
                    "setVar( codeBlock , compile( '1' ) )",
                    runEnv1,
                    parse("1", null));
            }
            catch (Exception e)
            {
                e.printStackTrace();
                throw new RuntimeException("" + e);
            }
            testEval("eval( codeBlock )", runEnv1, new BigDecimal(1));

            testEval("isValid('1+1')", Boolean.TRUE);

            testEval("isValid('1+?++1')", Boolean.FALSE);

            testEval("isValid('1+++2')", Boolean.FALSE);

            testParseException(" 4++", 1);

            testEval("true && ( 1 > 2)", Boolean.FALSE);

            testEval(
                "{setVarForName( 'testStrForName' , 'Inhalt testStrForName' ) ; testStrForName }",
                "Inhalt testStrForName");
            testEval("( 1 + ( 3 * 3 ) + 4 ) * 10", new BigDecimal(140));

            testEval(
                "{ println( 1 ) ;  println( 2 ) ;  println( 3 ) ;  println( 4 ) ;  println( 5 ) ;  println( 6 ) }",
                new BigDecimal(6));

            testEval(
                "{ setVar( i , 0 ) ; { println( 'sequ:'+i++ ) ; if( i < 3 , contSequ() , breakSequ( i ) ) } }",
                new BigDecimal(3));

            testEval(
                "{ setVar( i , 0 ) ; while( i < 3 , { if( i < 1 , println( 'i ist kleiner als 1' ) , if( i < 2 , println( 'i ist kleiner als 2' ) , println( 'i ist groesser gleich 2' ) ) ) ; i++ } ) }",
                new BigDecimal(2));

            testEval(
                "for( setVar( i , 0 ) , i < 10 , i++ , println( 'for:'+i ) )",
                "for:9");
            testEval(
                "for( setVar( i , 0 ) , i < 10 , null , println( 'for:'+i++ ) )",
                "for:9");

            testEval(
                "{ setVar( arr1 , newStringArray( 1 ) ) ; setVar( arr1[0] , 'str0' ) ; arr1[0] }",
                "str0");

            testEval(
                "{ setVar( arr1 , newStringArray( 'str0' , 'str1' ) ) ; arr1[0]+arr1[1] }",
                "str0str1");

            testEval("{ setVar( var1 , 1 ) ; removeVar( var1 ) ; var1 }", null);

            testEval(
                "{ setVarForName( 'var1' , 1 ) ; removeVarForName( 'var1' ) ; var1 }",
                null);

            testEval("num( '3' ) + 3", new BigDecimal(6));

            testEval("type( 'a' )", "STRING");

            testEval("type( 1 )", "NUMBER");

            testEval("type( null )", "NULL");

            testEval("type( true )", "BOOLEAN");

            testEval("type( compile( '1' ) )", "EXPRESSION");

            testEval("type( newStringArray( 1 ) )", "STRINGARRAY");

            testEval("isArray( newStringArray( 1 ) )", Boolean.TRUE);

            testEval("isArray( 1 )", Boolean.FALSE);

            testEval(
                "trycatch( 1 + varNotExist , 'gescheitert' )",
                "gescheitert");

            testEval("{}", null);

            testEval(
                "{ println( 1 ) ;  println( 2 ) ;  println( 3 ) ;  println( 4 ) ;  println( 5 ) ;  println( 6 ) }",
                new BigDecimal(6));

            testEval(
                "{ setVar( i , 0 ) ; { println( 'sequ:'+i++ ) ; if( i < 3 , contSequ() , breakSequ( i ) ) } }",
                new BigDecimal(3));

            testEval(
                "{ setVar( i , 0 ) ; while( i < 3 , { if( i < 1 , println( 'i ist kleiner als 1' ) , if( i < 2 , println( 'i ist kleiner als 2' ) , println( 'i ist groesser gleich 2' ) ) ) ; i++ } ) }",
                new BigDecimal(2));

            testEval(
                "{ setVar( x , 0 ) ; repeat( x > 0 , { println( 'tu solange x > 0, aber mindestens einmal' ) ; x-- } ) }",
                new BigDecimal(0));

            testEval(
                "{ setVar( exp1 , compile( highComma+'on the fly'+highComma ) ) ; eval( exp1 ) }",
                "on the fly");

            testEval(
                "{ setVar( arrx , newStringArray( 's0' , 's1' , 's2' ) ) ; println( arrx[ 0 ] ) ; println( arrx[ 1 ] ) ; println( arrx[ 2 ] ) }",
                "s2");

            //Test Kommentare
            testEval(" 1 // 1", new BigDecimal(1));

            testEval(" 1 + // 1\n 1", new BigDecimal(2));

            testEval(" 1 + /" + "* 1 *" + "/ 1", new BigDecimal(2));

            testEval(" 1 + /" + "** 1 *" + "/ 1", new BigDecimal(2));

            testEval(" 1 + /" + "**" + "/ 1", new BigDecimal(2));

            testParseException(" 1 + /" + "*" + "/ 1", 10);

            testEval(" 1 + /" + "**" + "/ 1", new BigDecimal(2));

            testEval(" varExists( existsNot )", Boolean.FALSE);

            // Laufzeit-Umgebung mit Variablen
            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("exists", Boolean.TRUE);
            testEval("varExists( exists )", runEnv1, Boolean.TRUE);

            testEval(" b1 := true", runEnv1, Boolean.TRUE);

            // Laufzeit-Umgebung mit Variablen
            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(1));
            testEval(" n1 += 1", runEnv1, new BigDecimal(2));

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("s1", "str1");
            testEval(" s1 += 'str2'", runEnv1, "str1str2");

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(1));
            testEval(" n1 -= 1", runEnv1, new BigDecimal(0));

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(2));
            testEval(" n1 *= 2", runEnv1, new BigDecimal(4));

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(4));
            testEval(" n1 /= 2", runEnv1, new BigDecimal(2));

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(4));
            testEval(
                " n1 :>= 2 // Zuweisung wenn kleiner",
                runEnv1,
                new BigDecimal(2));

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(4));
            testEval(
                " n1 :min= 2 // Zuweisung wenn kleiner",
                runEnv1,
                new BigDecimal(2));

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(2));
            testEval(
                " n1 :<= 4 // Zuweisung wenn groesser",
                runEnv1,
                new BigDecimal(4));

            runEnv1 = new StandaloneRuntimeEnvironment();
            runEnv1.setVariable("n1", new BigDecimal(2));
            testEval(
                " n1 :max= 4 // Zuweisung wenn groesser",
                runEnv1,
                new BigDecimal(4));

            //das funktioniert nur auf einem schnellen Rechner Differenz unter 1 Millisekunde
            //testEval("newObject( 'java.util.Date' )", new Date()) ;

            //das funktioniert nur auf einem schnellen Rechner Differenz unter 1 Millisekunde
            //testEval("date1:=newObject( 'java.util.Date' )", new Date()) ;

            testEval("a1:='a'", "a");

            testEval("n1:=null", null);

            testEval("pi", new BigDecimal(Math.PI));
            testEval("PI", new BigDecimal(Math.PI));

            testEval("euler", new BigDecimal(Math.E));
            testEval("Euler", new BigDecimal(Math.E));
            testEval("EULER", new BigDecimal(Math.E));

            testEval("sin( 0 )", new BigDecimal(0));

            testEval("cos( 0 )", new BigDecimal(1));

            testEval("abs( -1 )", new BigDecimal(1));

            testEval("acos( 1 )", new BigDecimal(0));

            testEval("asin( 0 )", new BigDecimal(0));

            testEval("atan( 0 )", new BigDecimal(0));

            testEval("ceil( 0.5 )", new BigDecimal(1));

            testEval("exp( 0 )", new BigDecimal(1));

            testEval("floor( 1.1 )", new BigDecimal(1));

            testEval("log( 1 )", new BigDecimal(0));

            testEval("pow( 2 , 2 )", new BigDecimal(4));

            testEval("println( random() ) == 2", Boolean.FALSE);

            testEval("tan( 0 )", new BigDecimal(0));

            testEval("toDegrees( pi * 2 )", new BigDecimal(360));

            testEval("toRadians( 90 )", new BigDecimal(Math.PI / 2));

            // pi ist ein reserviertes Wort und kann nicht zugewiesen werden
            testEvalException("setVarForName( 'pi' , 10 )", runEnv1, 0);

            // Pi ist ein reserviertes Wort und kann nicht zugewiesen werden
            testEvalException("setVarForName( 'PI' , 10 )", runEnv1, 0);

            // Euler ist ein reserviertes Wort und kann nicht zugewiesen werden
            testEvalException("setVarForName( 'Euler' , 10 )", runEnv1, 0);

            testEval("1 + 2 * 3", new BigDecimal(7)); // Test Operator-Priority

            // Test Operator-Priority
            testEval("2 * 2 + 2 * 3", new BigDecimal(10));

            testEval("(2) * 2 + 2 * 3", new BigDecimal(10));

            // Test Operator-Priority
            testEval(" 10 == ( 2 * 2 + 2 * 3 )", Boolean.TRUE);

            // Test Operator-Priority
            testEval(" 10 == 2 * 2 + 2 * 3 ", Boolean.TRUE);

            // Test Operator-Priority
            testEval(" 10 + 1 == 2 * 2 + 2 * 3 + 1 ", Boolean.TRUE);

            // Test korrigierte Fehlermeldung mit Beachtung Zeilenumbrche (computed Line Col)
            testParseException(
                " 1 + // 1\n 1 + 'zahl + string geht nicht'",
                13);

            // Test korrigierte Fehlermeldung mit Beachtung Zeilenumbrche (computed Line Col)
            testEvalException(" 1 + // 1\n 1 + varExistsNicht", runEnv1, 13);

            // auch leere Anweisungen in Sequencen erlaubt
            testEval("{;;;}", null);

            //indizierter Zugriff auf Maps
            Map testMap = new HashMap();

            testMap.put("A", "varMap 'A'");
            testMap.put("B", "varMap 'B'");
            testMap.put("C", "varMap 'C'");

            runEnv1.setVariable("varMap", testMap);
            testEval("varMap[ 'A' ]", runEnv1, "varMap 'A'");
            testEval("varMap[ 'B' ]", runEnv1, "varMap 'B'");
            testEval("varMap[ 'C' ]", runEnv1, "varMap 'C'");

            //mehrdimensionale Zugriffe
            String[][] strArrArr = new String[][] { { "A1", "A2" }, {
                    "B1", "B2" }, };

            runEnv1.setVariable("strArrArr", strArrArr);

            testEval("strArrArr[ 0 ][ 0 ]", runEnv1, "A1");

            testEval("strArrArr[ 0 ][ 0 ] := 'Test'", runEnv1, "Test");

            testEval("strArrArr[ 0 ][ 0 ]", runEnv1, "Test");

            // Test Array, List und Map geschachtelt
            Object[] objArr = new Object[1];

            ArrayList arrList = new ArrayList();

            objArr[0] = arrList;

            HashMap hashMap = new HashMap();

            arrList.add(hashMap);

            hashMap.put("A", "A");

            runEnv1.setVariable("arrListMap", objArr);

            testEval("arrListMap[ 0 ][ 0 ][ 'A' ]", runEnv1, "A");

            testEval("arrListMap[ 0 , 0 , 'A' ]", runEnv1, "A");

            testEval("arrListMap[ 0 ][ 0 ][ 'A' ] := 'Test'", runEnv1, "Test");

            testEval("arrListMap[ 0 , 0 , 'A' ]", runEnv1, "Test");

            testEval("arrListMap[ 0 , 0 , 'A' ] := 'Test1'", runEnv1, "Test1");

            testEval("arrListMap[ 0 , 0 , 'A' ]", runEnv1, "Test1");

            testEval("( arrListMap )[ 0 , 0 , 'A' ]", runEnv1, "Test1");

            testEval("removeVar( arrListMap[ 0 , 0 , 'A' ] )", runEnv1, null); // "Test1"); ??? @todo: sollte wert zurckliefern

            testEval("arrListMap[ 0 , 0 , 'A' ]", runEnv1, null);

            runEnv1.setVariable(
                "bigArr",
                new BigDecimal[] { new BigDecimal(1)});
            testEval("bigArr[ 0 ]++", runEnv1, new BigDecimal(1));
            testEval("bigArr[ 0 ]++", runEnv1, new BigDecimal(2));

            runEnv1.setVariable("intArr", new Integer[] { new Integer(1) } );
            testEval("intArr[ 0 ]++", runEnv1, new Integer(1));
            testEval("intArr[ 0 ]", runEnv1, new Integer(2));

            //Array als Ergebnis eines Funktionsaufrufes, von dem wiederum ein Element adressiert wird
            testEval("newStringArray( 'str0' , 'str1' )[ 0 ]", "str0");
            testEval("( newStringArray( 'str0' , 'str1' ) )[ 0 ]", "str0");
            testEval("{ newStringArray( 'str0' , 'str1' ) }[ 0 ]", "str0");

            // Test CnC-Templates
            testEval("eval( compileTemplate( \"aaa <%= 'bbb' %> ccc\" ) )", runEnv1, "aaa bbb ccc");

            testEval("eval( compileTemplate( \"aaa <% for( i := 0 , i <= 3 , i++ , { %>bbb <% } ) ;%>ccc\" ) )", runEnv1, "aaa bbb bbb bbb bbb ccc");

            testEval("eval( compileTemplate( \"a <% for( i1 := 0 , i1 <= 3 , i1++ , { for( i2 := 0 , i2 <= 3 , i2++ , { %>b<% } ) ; { %>_<% } } ) ;%> c\" ) )", runEnv1, "a bbbb_bbbb_bbbb_bbbb_ c");

            testEval("eval( compileTemplate( 'a <% for( i1 := 0 , i1 <= 3 , i1++ , { %>b<% for( i2 := 0 , i2 <= 3 , i2++ , { %>c<% } ) ;%> <% } ) ;%> d' ) )", runEnv1, "a bcccc bcccc bcccc bcccc  d");

            Test test = new Test();
            test.tstStr = "tstStr value";
            runEnv1.setVariable("testMember", test);
            testEval("testMember.tstStr", runEnv1, "tstStr value");
            testEval("testMember.tstStr:='tstStr value new'", runEnv1, "tstStr value new");
            testEval("testMember.tstStr", runEnv1, "tstStr value new");

            testEval("testMember .tstStr", runEnv1, "tstStr value new");

            testEval("testMember. tstStr", runEnv1, "tstStr value new");

            testEval("testMember . tstStr", runEnv1, "tstStr value new");

            testEval("(1).toString()", "1" );

            testEval("1.toString()", "1" );

            testEval("testMember.setTstStr( 'Xyz' )", runEnv1, null);

            testEval("testMember.tstStr", runEnv1, "Xyz");

            testEval("testMember.setTstStr( null )", runEnv1, null);

            testEval("testMember.setTstStr( 'Xyz1' , 3 , true )", runEnv1, null);

            testEval("testMember.tstStr", runEnv1, "Xyz1");

            testEval("testMember.setTstStr( 'Xyz2' , true , 3 )", runEnv1, null);

            testEval("testMember.tstStr", runEnv1, "Xyz2");

            testEval("arrListMap[ 0 ][ 0 ].remove( 'A' )", runEnv1, null);

            testEval("arrListMap[ 0 , 0 , 'A' ]", runEnv1, null);

            // Doppel-Punkt im Variablen-Name
            runEnv1.setVariable("a:a", new BigDecimal(1));
            testEval("a:a", runEnv1, new BigDecimal(1));

            //Funktion timeOfDay
            Date date = new Date();
            runEnv1.setVariable("date", date);
            //Umwandlung Time -> String -> Time, weil sonst nicht korrekt Comparable (Tagesdatum existiert noch im long)
            testEval("timeOfDay( date )", runEnv1, Time.valueOf( new Time( date.getTime() ).toString() ) );

            //Funktion stripTimeOfDay
            testEval("stripTimeOfDay( date )", runEnv1, new java.sql.Date( date.getTime() ) );

            //Test Time-Literal
            testEval("00:01:01", runEnv1, new java.sql.Time( 0 , 1 , 1 ) );
            testEval("01:01:01", runEnv1, new java.sql.Time( 1 , 1 , 1 ) );
            testEval("18:31", runEnv1, new java.sql.Time( 18 , 31 , 0 ) );

            //Test dayOfWeek
            date = new Date( 105 , 03 , 11  ); // 2005-04-11, Montag
            runEnv1.setVariable("date", date);
            testEval("date", runEnv1, date );
            testEval("dayOfWeek( date )", runEnv1, new Integer( Calendar.MONDAY ) );
            
            //Test Zeit-Bereich
            date = new Date( 105 , 03 , 11 , 16 , 30 , 0 ); // 2005-04-11, Montag 16:30
            System.out.println( date );
            runEnv1.setVariable("date", date);
            testEval("date", runEnv1, date );


            //runEnv1.setVariable("SUNDAY"    , new Integer( Calendar.SUNDAY ) );
            //runEnv1.setVariable("MONDAY"    , new Integer( Calendar.MONDAY ) );
            //runEnv1.setVariable("TUESDAY"   , new Integer( Calendar.TUESDAY ) );
            //runEnv1.setVariable("WEDNESDAY" , new Integer( Calendar.WEDNESDAY ) );
            //runEnv1.setVariable("THURSDAY"  , new Integer( Calendar.THURSDAY ) );
            //runEnv1.setVariable("FRIDAY"    , new Integer( Calendar.FRIDAY ) );
            //runEnv1.setVariable("SATURDAY"  , new Integer( Calendar.SATURDAY ) );

            //Test mit den Funktionen dayOfWeek und timeOfDay
            testEval("( ( dayOfWeek( date ) >= MONDAY ) && ( dayOfWeek( date ) <= FRIDAY ) )", runEnv1, Boolean.TRUE );
            testEval("'' + timeOfDay( date )", runEnv1, "16:30:00" );
            testEval("'' + 15:00", "15:00:00" );
            testEval("'' + 18:00", "18:00:00" );
            testEval("( timeOfDay( date ) >= 15:00 )", runEnv1, Boolean.TRUE );
            testEval("15:00 < 18:00", runEnv1, Boolean.TRUE );
            testEval("timeOfDay( date ) <= 18:00", runEnv1, Boolean.TRUE );
            testEval("( ( timeOfDay( date ) >= 15:00 ) && ( timeOfDay( date ) <= 18:00 ) )", runEnv1, Boolean.TRUE );
            testEval("( ( dayOfWeek( date ) >= MONDAY ) && ( dayOfWeek( date ) <= FRIDAY ) ) && ( ( timeOfDay( date ) >= 15:00 ) && ( timeOfDay( date ) <= 18:00 ) )", runEnv1, Boolean.TRUE );
            testEval("dayOfWeek( date ) >= MONDAY && dayOfWeek( date ) <= FRIDAY && timeOfDay( date ) >= 15:00 && timeOfDay( date ) <= 18:00", runEnv1, Boolean.TRUE );
            testEval("( dayOfWeek( date ) >= MONDAY ) && ( dayOfWeek( date ) <= FRIDAY ) && ( ( ( timeOfDay( date ) >= 6:30 ) && ( timeOfDay( date ) <= 9:30 ) ) || ( ( timeOfDay( date ) >= 15:00 ) && ( timeOfDay( date ) <= 18:00 ) ) )", runEnv1, Boolean.TRUE );
            testEval("dayOfWeek( date ) >= MONDAY && dayOfWeek( date ) <= FRIDAY && ( ( timeOfDay( date ) >= 6:30 && timeOfDay( date ) <= 9:30 ) || ( timeOfDay( date ) >= 15:00 && timeOfDay( date ) <= 18:00 ) )", runEnv1, Boolean.TRUE );

            //Test ohne die Funktionen dayOfWeek und timeOfDay
            testEval("( ( date >= MONDAY ) && ( date <= FRIDAY ) )", runEnv1, Boolean.TRUE );
            testEval("( date >= 15:00 )", runEnv1, Boolean.TRUE );
            testEval("15:00 < 18:00", runEnv1, Boolean.TRUE );
            testEval("date <= 18:00", runEnv1, Boolean.TRUE );
            testEval("( ( date >= 15:00 ) && ( date <= 18:00 ) )", runEnv1, Boolean.TRUE );
            testEval("( ( date >= MONDAY ) && ( date <= FRIDAY ) ) && ( ( date >= 15:00 ) && ( date <= 18:00 ) )", runEnv1, Boolean.TRUE );
            testEval("date >= MONDAY && date <= FRIDAY && date >= 15:00 && date <= 18:00", runEnv1, Boolean.TRUE );
            testEval("( date >= MONDAY ) && ( date <= FRIDAY ) && ( ( ( date >= 6:30 ) && ( date <= 9:30 ) ) || ( ( date >= 15:00 ) && ( date <= 18:00 ) ) )", runEnv1, Boolean.TRUE );
            testEval("date >= MONDAY && date <= FRIDAY && ( ( date >= 6:30 && date <= 9:30 ) || ( date >= 15:00 && date <= 18:00 ) )", runEnv1, Boolean.TRUE );

            //Test Monate
            Timestamp timestamp = Timestamp.valueOf("2005-05-01 00:00:00");
            runEnv1.setVariable("timestamp",timestamp);
            testEval("timestamp", runEnv1, timestamp);
            testEval("MAY", runEnv1, Month.MAY);
            testEval("timestamp == MAY", runEnv1, Boolean.TRUE);
            testEval("timestamp >= MARCH", runEnv1, Boolean.TRUE);
            testEval("timestamp <= JUNE", runEnv1, Boolean.TRUE);
            testEval("month(timestamp) == MAY", runEnv1, Boolean.TRUE);

            //Test gebrochene Sekunden
            
            //Time time = new Time( Timestamp.valueOf( "1970-01-01 15:00:00.23" ).getTime() );
            //Time time = new Time( Math.round( ( ( ( ( 14 * 60 ) + 0 ) * 60 ) + 0.23 ) * 1000 ) );
            Time time = TimeOfDayLiteral.addFractionalMillis( new Time( 15 , 0 , 0 ) , 0.23 );
            runEnv1.setVariable("time",time);
            testEval("15:00:00.23", runEnv1, time);
			testEval("time == 15:00:00.23", runEnv1, Boolean.TRUE);
            
            //time = new Time( Timestamp.valueOf( "1970-01-01 18:00:00.34" ).getTime() );
            //time = new Time( Math.round( ( ( ( ( 17 * 60 ) + 0 ) * 60 ) + 0.34 ) * 1000 ) );
            time = TimeOfDayLiteral.addFractionalMillis( new Time( 18 , 0 , 0 ) , 0.34 );
            runEnv1.setVariable("time",time);
            testEval("time == 18:00:00.34", runEnv1, Boolean.TRUE);

            //Test Zeit-Bereich
            date = new Date( 105 , 04 , 6 , 7 , 37 , 38 ); // 2005-05-06 07:37:38.0 Freitag

            System.out.println( StringUtil.toFullStr( date ) );
            runEnv1.setVariable("date", date);
            testEval("date", runEnv1, date );
            testEval("((date == MONDAY) || ((date == TUESDAY) && (date >= 06:00 && date <= 12:00)) || ((date == WEDNESDAY) && (date >= 11:00 && date <= 15:00)) || (date == THURSDAY) || ((date == FRIDAY) && (date >= 07:00 && date <= 08:00 )))", runEnv1, Boolean.TRUE );

//TODO date time literale            
//TODO Functions als Plugins
            
            //Ausgabe der aktuellen TimeZone
            //Calendar calendar = new GregorianCalendar();
            //System.out.println(calendar.getTimeZone());

            //neue Notation fr Datum und Uhrzeit noch nicht fertig
            //testEval("2005T", runEnv1, new Year( 2005 ) );
            //testEval("2005-05T", runEnv1, date );
            //testEval("2005-05-06T", runEnv1, date );
            //testEval("2005-05-06T07:37", runEnv1, date );
            //testEval("2005-05-06T07:37:38", runEnv1, date );

            //debugStop = true;

            //@todo: Test add(), Test empty()

            // testEval("", ) ;

            // Punkt im Variablen-Name
            //runEnv1.setVariable("a.a", new BigDecimal(1));
            //testEval("a.a", runEnv1, new BigDecimal(1));

            // Typ-Prfung
            runEnv1 = new StandaloneRuntimeEnvironment();
            // Definiere 'str' als String
            testEval("var java.lang.String str", runEnv1, String.class);
            // Zuweisen Nummer sollte fehlschlagen
            testEvalException("str := 1", runEnv1, 0);

            // Kontrolle auf Property
            testEval("var de.cnc.expression.Test test", runEnv1, Test.class);
            //testEval("test := newObject('de.cnc.expression.Test')", runEnv1, new Test() );
            runEnv1.setVariable("test", new Test() );

            
            // Ansprechen Property sollte fehlschlagen
            testParseException("test.propNoExist", runEnv1, 4);

            // Aufruf Methode sollte fehlschlagen
            testParseException("test.methodNoExist()", runEnv1, 4);

            System.out.println("\n--- Test war ok ---");
        }
        catch (ExpressionEvaluationException exEvalExc)
        {
            System.err.println(exEvalExc);
            exEvalExc.printStackTrace();
        }

    }

}
