package de.cnc.expression.tokencomplex;

import java.lang.reflect.Field;
import java.lang.reflect.Method;

import de.cnc.expression.AbstractRuntimeEnvironment;
import de.cnc.expression.AbstractToken;
import de.cnc.expression.Assignable;
import de.cnc.expression.Types;
import de.cnc.expression.exceptions.ExpressionEvaluationException;
import de.cnc.expression.exceptions.ExpressionParseException;
import de.cnc.expression.util.ParserUtil;
import de.cnc.util.BeanUtil;

/**
 * Member-Dot-Operator <b>&lt;index&gt;.member</b>  zur 
 * Adressierung eines Members eines Objekts . <br>
 *
 * <br/>
 * <b>Author:</b> <a href="http://www.heinerkuecker.de" target="_blank">Heiner Kcker</a><br/>
 * <br/>
 *
 * @author Heiner Kcker
 * @version $Id:$
 */
public class MemberDotOperator extends AbstractDotOperator implements Assignable
{
    /**
     * der Ausdruck, hinter dem ein Dot-Operator steht
     */
    //AbstractToken dottedExpression;
    
    /** Name des adressierten Objekt-Members */
    String memberName;

    /**
     *
     */
    public static MemberDotOperator parseMemberDotOperator(
        String paStr,
        int paIntLine,
        int paIntCol,
        String paStrOriginalSource)
        //throws ExpressionParseException
    { // , NoSuchTokenException {
//System.out.println("MemberDotOperator parseMemberDotOperator: " + paStr);
        if (!paStr.startsWith(".")) {
            return null;
        }

        int iWhites = ParserUtil.parseWhitesAndComments( paStr.substring( 1 ) );

        final int identifierLen = ParserUtil.parseIdentifier( paStr.substring( 1 + iWhites )  );
        if (identifierLen < 1)
        {
            return null;
        }
        
        final MemberDotOperator dotElem = new MemberDotOperator(paStrOriginalSource, paIntLine, paIntCol);
        dotElem.memberName = paStr.substring(iWhites + 1, iWhites + identifierLen + 1);
        int iRun = 1 + iWhites + identifierLen;

        dotElem.strSource = paStr.substring(0, iRun);
        dotElem.iSrcLength = iRun;

        return dotElem;

    } // end method parse

    /**
     * constructor
     */
    public MemberDotOperator(
        String paStrOriginalSource,
        int paIntLine,
        int paIntCol)
    {
        super(paStrOriginalSource, paIntLine, paIntCol);
        // System.out.println( "constructor paStrOriginalSource:"+paStrOriginalSource);
    } // end constructor

    /**
     * set the dotted expression.
     * Setzt den Ausdruck, der am Ende eine Index-Element-Operator []
     * trgt. Somit wird sozusagen der hinter dem indizierten Ausdruck stehende
     * Index-Operator vor den indizierten Ausdruck gesetzt.
     * Demzufolge wird erst der indizierte Ausdruck evaluiert und auf dessen Ergebnis
     * dann der Index-Operator angewendet.
     * 
     * @param token
     */
    public void setDottedExpression(AbstractToken pToken)
    {
        this.dottedExpression = pToken;
        //aufaddieren der Lnge des gedotteten Ausdruckes,
        //da gilt >xxx[ yyy ]< Lnge indizierter Ausdruck + Lnge Index-Operator
        this.iSrcLength += pToken.getSrcLength();
    }

   /**
     * rekursives Aufrufen der Klammerung der Ausdrcke
     * entsprechend der Operator-Prioritt.
     * Weiterleiten an den Index-Auisdruck ( var[ &lt;exp&gt; ] )
     */
    public final void organizeOperatorPriority()
    {
        this.dottedExpression.organizeOperatorPriority();
    }

    /**
     * Es wird erst der gedottete Ausdruck evaluiert und auf dessen Ergebnis
     * dann der Dot-Operator angewendet.
     */
    public final Object eval(final AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionEvaluationException
    {
// TODO spezial-mimik fr .length auf ein Array oder List
        if (! pRunEnv.isMemberEnabled() )
        {
            throw new ExpressionEvaluationException(
                    getLine(),
                    getCol(),
                    getSource(),
                    getOriginalSource(),
                    "member access not enabled");
        }
        Object dottedObj = this.dottedExpression.eval(pRunEnv);
        //System.out.println( "eval collObj: " + collObj );

        if ( dottedObj == null )
        {
            throw new ExpressionEvaluationException(
                    this.dottedExpression.getLine(),
                    this.dottedExpression.getCol(),
                    this.dottedExpression.getSource(),
                    this.dottedExpression.getOriginalSource(),
                    this.dottedExpression.getSource() + " is null");
        }
        
        return  BeanUtil.get(dottedObj, this.memberName );
    }

    /**
     * debug output
     */
    public String toString()
    {
        String strRet =
            getClass().getName()
                + " srcLength: "
                + getSrcLength()
                + " source: "
                + getSource()
                + " dotted expression: "
                + this.dottedExpression;

        return strRet;
    }

    /**
     * set value unter beachtung array index
     */
    public void set(//--
            final AbstractRuntimeEnvironment pRunEnv,
            final Object pObjValue) // zu setzender Wert
        throws ExpressionEvaluationException
    {
        if (! pRunEnv.isMemberEnabled() )
        {
            throw new ExpressionEvaluationException(
                    getLine(),
                    getCol(),
                    getSource(),
                    getOriginalSource(),
                    "member access not enabled");
        }

        // Holen gedottetes Objekt
       Object dottedObj = this.dottedExpression.eval( pRunEnv );
       BeanUtil.set(dottedObj, this.memberName, pObjValue);
    }

    /**
     * remove on object member not supported
     */
    public Object remove(AbstractRuntimeEnvironment pRunEnv)
        throws ExpressionEvaluationException
    {
                    throw new ExpressionEvaluationException(
                        getLine(),
                        getCol(),
                        getSource(),
                        getOriginalSource(),
                        "remove on object member not supported");
    } 


    /**
     * get the wahrscheinlichen return typen .<br>
     * Wahrscheinlicher Return Type zum vorbeugenden Prfen der
     * Funktionsparameter null wenn kein Rckgabe-Wert, mehrere wenn mehrere
     * mglich ( z.B. Plus-Operator kann numerisch oder String liefern)
     */
    public final Object[] getCheckReturnTypes(//--
            final AbstractRuntimeEnvironment pRunEnv)
            throws ExpressionParseException
    {
//      TODO spezial-mimik fr .length auf ein Array oder List
        final Object[] typArr = //--
            this.dottedExpression
                .getCheckReturnTypes(pRunEnv);

        //Log.debug( "memberName" , this.memberName );
        //Log.debug( "typArr" , typArr );
        if (typArr != null //--
                && typArr.length > 0)
        {
            Object typ = typArr[0];
            //Log.debug( "typ" , typ );
            if (typ instanceof Class)
            {
                Class clazz = (Class) typ;
                //Log.debug( "clazz" , clazz );
                Field field = BeanUtil.getFieldCatched(clazz, this.memberName);
                //Log.debug("field", field);
                if (field != null)
                {
                    return new Object[]
                    {field.getType()};
                }
                Method method = BeanUtil.getGetMethodCatched(clazz, this.memberName);
                //Log.debug("method", method);
                if (method != null)
                {
                    return new Object[]
                    {method.getReturnType()};
                }
                else // if (method == null)
                {
                    throw new ExpressionParseException(getLine(), getCol(),
                            getSource(), "no field " + this.memberName);
                }
            }
        }
        return new Object[]{Types.UNKNOWN_CLASS};
    } // end method getReturnTypes

    /* (non-Javadoc)
     * @see de.cnc.expression.AbstractToken#isConst(de.cnc.expression.AbstractRuntimeEnvironment)
     */
    public final boolean isConst(final AbstractRuntimeEnvironment pRunEnv) throws ExpressionEvaluationException
    {
        return this.dottedExpression.isConst(pRunEnv);
    }

    /* (non-Javadoc)
     * @see de.cnc.expression.AbstractToken#getConstVal(de.cnc.expression.AbstractRuntimeEnvironment)
     */
    public final Object getConstVal(final AbstractRuntimeEnvironment pRunEnv) throws ExpressionEvaluationException
    {
        return eval(pRunEnv);
    }

}
